from __future__ import annotations

from ai.backend.client.request import Request
from ai.backend.common.container_registry import (
    PatchContainerRegistryRequestModel,
    PatchContainerRegistryResponseModel,
)

from .base import BaseFunction, api_function

__all__ = ("ContainerRegistry",)


class ContainerRegistry(BaseFunction):
    """
    Provides functions to manage container registries.
    """

    @api_function
    @classmethod
    async def patch_container_registry(
        cls, registry_id: str, params: PatchContainerRegistryRequestModel
    ) -> PatchContainerRegistryResponseModel:
        """
        Updates the container registry information, and return the container registry.

        :param registry_id: ID of the container registry.
        :param params: Parameters to update the container registry.
        """

        request = Request(
            "PATCH",
            f"/container-registries/{registry_id}",
        )
        request.set_json(params)

        async with request.fetch() as resp:
            return await resp.json()



from typing import Any, Sequence

from ai.backend.common.types import QuotaConfig, QuotaScopeID

from ..output.fields import group_fields, quota_scope_fields, user_fields
from ..output.types import FieldSpec
from ..session import api_session
from ..types import set_if_set
from ..utils import dedent as _d
from .base import BaseFunction, api_function

_default_user_fields = (
    user_fields["uuid"],
    user_fields["username"],
)

_default_project_fields = (
    group_fields["id"],
    group_fields["name"],
)

_default_detail_fields = (
    quota_scope_fields["usage_bytes"],
    quota_scope_fields["usage_count"],
    quota_scope_fields["hard_limit_bytes"],
)

_default_quota_scope_fields = (
    quota_scope_fields["quota_scope_id"],
    quota_scope_fields["storage_host_name"],
)


class QuotaScope(BaseFunction):
    @api_function
    @classmethod
    async def get_user_info(
        cls,
        domain_name: str,
        email: str,
        fields: Sequence[FieldSpec] = _default_user_fields,
    ) -> dict[str, Any]:
        query = _d("""
            query($domain_name: String!, $email: String!) {
                user(domain_name: $domain_name, email: $email) { $fields }
            }
        """)
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {
            "domain_name": domain_name,
            "email": email,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["user"]

    @api_function
    @classmethod
    async def get_project_info(
        cls,
        domain_name: str,
        name: str,
        fields: Sequence[FieldSpec] = _default_project_fields,
    ) -> dict[str, Any]:
        query = _d("""
            query($domain_name: String!, $name: String!) {
                groups_by_name(domain_name: $domain_name, name: $name) { $fields }
            }
        """)
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {
            "domain_name": domain_name,
            "name": name,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["groups_by_name"][0]

    @api_function
    @classmethod
    async def get_quota_scope(
        cls,
        host: str,
        qsid: QuotaScopeID,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict[str, Any]:
        query = _d("""
            query($storage_host_name: String!, $quota_scope_id: String!) {
                quota_scope(storage_host_name: $storage_host_name, quota_scope_id: $quota_scope_id) {
                    details { $fields }
                }
            }
        """)
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {
            "storage_host_name": host,
            "quota_scope_id": str(qsid),
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["quota_scope"]["details"]

    @api_function
    @classmethod
    async def set_quota_scope(
        cls,
        host: str,
        qsid: QuotaScopeID,
        config: QuotaConfig,
        fields: Sequence[FieldSpec] = _default_quota_scope_fields,
    ) -> dict[str, Any]:
        query = _d("""
            mutation($storage_host_name: String!, $quota_scope_id: String!, $input: QuotaScopeInput!) {
                set_quota_scope(storage_host_name: $storage_host_name, quota_scope_id: $quota_scope_id, props: $input) {
                    quota_scope { $fields }
                }
            }
        """)
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        inputs: dict[str, Any] = {}
        set_if_set(inputs, "hard_limit_bytes", config.limit_bytes)
        variables = {
            "storage_host_name": host,
            "quota_scope_id": str(qsid),
            "input": inputs,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["set_quota_scope"]["quota_scope"]

    @api_function
    @classmethod
    async def unset_quota_scope(
        cls,
        host: str,
        qsid: QuotaScopeID,
        fields: Sequence[FieldSpec] = _default_quota_scope_fields,
    ) -> dict[str, Any]:
        query = _d("""
            mutation($storage_host_name: String!, $quota_scope_id: String!) {
                unset_quota_scope(storage_host_name: $storage_host_name, quota_scope_id: $quota_scope_id) {
                    quota_scope { $fields }
                }
            }
        """)
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {
            "storage_host_name": host,
            "quota_scope_id": str(qsid),
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["unset_quota_scope"]["quota_scope"]



from __future__ import annotations

import asyncio
import enum
import functools
import json
import sys
import textwrap
import traceback
from collections.abc import Sequence
from types import TracebackType
from typing import Optional, Self

from click import echo, style
from tqdm import tqdm

from ..exceptions import BackendAPIError

__all__ = (
    "PrintStatus",
    "print_pretty",
    "print_info",
    "print_wait",
    "print_done",
    "print_warn",
    "print_fail",
    "print_error",
    "show_warning",
)


class PrintStatus(enum.Enum):
    NONE = 0
    WAITING = 1
    DONE = 2
    FAILED = 3
    WARNING = 4


def bold(text: str) -> str:
    """
    Wraps the given text with bold enable/disable ANSI sequences.
    """
    return style(text, bold=True, reset=False) + style("", bold=False, reset=False)


def underline(text: str) -> str:
    return style(text, underline=True, reset=False) + style("", underline=False, reset=False)


def inverse(text: str) -> str:
    return style(text, reverse=True, reset=False) + style("", reverse=False, reset=False)


def italic(text: str) -> str:
    return "\x1b[3m" + text + "\x1b[23m"


def format_pretty(msg, status=PrintStatus.NONE, colored=True):
    if status == PrintStatus.NONE:
        indicator = style("\u2219", fg="bright_cyan", reset=False)
    elif status == PrintStatus.WAITING:
        indicator = style("\u22ef", fg="bright_yellow", reset=False)
    elif status == PrintStatus.DONE:
        indicator = style("\u2714", fg="bright_green", reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style("\u2718", fg="bright_red", reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style("\u2219", fg="yellow", reset=False)
    else:
        raise ValueError
    return style(indicator + textwrap.indent(msg, "  ")[1:], reset=True)


format_info = functools.partial(format_pretty, status=PrintStatus.NONE)
format_wait = functools.partial(format_pretty, status=PrintStatus.WAITING)
format_done = functools.partial(format_pretty, status=PrintStatus.DONE)
format_fail = functools.partial(format_pretty, status=PrintStatus.FAILED)
format_warn = functools.partial(format_pretty, status=PrintStatus.WARNING)


def print_pretty(msg, *, status=PrintStatus.NONE, file=None):
    if file is None:
        file = sys.stderr
    if status == PrintStatus.NONE:
        indicator = style("\u2219", fg="bright_cyan", reset=False)
    elif status == PrintStatus.WAITING:
        assert "\n" not in msg, "Waiting message must be a single line."
        indicator = style("\u22ef", fg="bright_yellow", reset=False)
    elif status == PrintStatus.DONE:
        indicator = style("\u2713", fg="bright_green", reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style("\u2717", fg="bright_red", reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style("\u2219", fg="yellow", reset=False)
    else:
        raise ValueError
    echo("\x1b[2K", nl=False, file=file)
    text = textwrap.indent(msg, "  ")
    text = style(indicator + text[1:], reset=True)
    echo("{0}\r".format(text), nl=False, file=file)
    file.flush()
    if status != PrintStatus.WAITING:
        echo("", file=file)


print_info = functools.partial(print_pretty, status=PrintStatus.NONE)
print_wait = functools.partial(print_pretty, status=PrintStatus.WAITING)
print_done = functools.partial(print_pretty, status=PrintStatus.DONE)
print_fail = functools.partial(print_pretty, status=PrintStatus.FAILED)
print_warn = functools.partial(print_pretty, status=PrintStatus.WARNING)


def _format_gql_path(items: Sequence[str | int]) -> str:
    pieces = []
    for item in items:
        match item:
            case int():
                pieces.append(f"[{item}]")
            case _:
                pieces.append(f".{str(item)}")
    return "".join(pieces)[1:]  # strip first dot


def format_error(exc: Exception):
    if isinstance(exc, BackendAPIError):
        yield "{0}: {1} {2}\n".format(exc.__class__.__name__, exc.status, exc.reason)
        yield "{0[title]}".format(exc.data)
        if exc.data["type"].endswith("/too-many-sessions-matched"):
            matches = exc.data["data"].get("matches", [])
            if matches:
                yield "\nCandidates (up to 10 recent entries):\n"
            for item in matches:
                yield f"- {item['id']} ({item['name']}, {item['status']})\n"
        elif exc.data["type"].endswith("/session-already-exists"):
            existing_session_id = exc.data["data"].get("existingSessionId", None)
            if existing_session_id is not None:
                yield f"\n- Existing session ID: {existing_session_id}"
        elif exc.data["type"].endswith("/invalid-api-params"):
            general_error_msg = exc.data.get("msg", None)
            if general_error_msg is not None:
                yield f"\n- {general_error_msg}"
            per_field_errors = exc.data.get("data", None)
            if per_field_errors:
                yield "\n"
                yield json.dumps(per_field_errors, indent=2)
        else:
            if exc.data["type"].endswith("/graphql-error"):
                yield "\n\u279c Message:\n"
                for err_item in exc.data.get("data", []):
                    yield f"{err_item['message']}"
                    if err_path := err_item.get("path"):
                        yield f" (path: {_format_gql_path(err_path)})"
                    yield "\n"
            else:
                other_details = exc.data.get("msg", None)
                if other_details:
                    yield "\n\u279c Message: "
                    yield str(other_details)
                other_data = exc.data.get("data", None)
                if other_data:
                    yield "\n\u279c Data: "
                    yield repr(other_data)
        agent_details = exc.data.get("agent-details", None)
        if agent_details is not None:
            yield "\n\u279c This is an agent-side error. "
            yield "Check the agent status or ask the administrator for help."
            agent_exc = agent_details.get("exception", None)
            if agent_exc is not None:
                yield "\n\u279c " + str(agent_exc)
            desc = agent_details.get("title", None)
            if desc is not None:
                yield "\n\u279c " + str(desc)
        content = exc.data.get("content", None)
        if content:
            yield "\n" + content
    else:
        args = exc.args if exc.args else [""]
        yield f"{exc.__class__.__name__}: {args[0]}\n"
        yield "\n".join(map(str, args[1:]))
        yield ("*** Traceback ***\n" + "".join(traceback.format_tb(exc.__traceback__)).strip())


def print_error(exc: Exception, *, file=None):
    if file is None:
        file = sys.stderr
    indicator = style("\u2718", fg="bright_red", reset=False)
    if file.isatty():
        echo("\x1b[2K", nl=False, file=file)
    text = "".join(format_error(exc))
    text = textwrap.indent(text, "  ")
    text = style(indicator + text[1:], reset=True)
    echo("{0}\r".format(text), nl=False, file=file)
    echo("", file=file)
    file.flush()


def show_warning(message, category, filename, lineno, file=None, line=None):
    echo(
        "{0}: {1}".format(
            style(str(category.__name__), fg="yellow", bold=True),
            style(str(message), fg="yellow"),
        ),
        file=file,
    )


class ProgressBarWithSpinner(tqdm):
    """
    A simple extension to tqdm adding a spinner.

    .. code-block::

       async with ProgressBarWithSpinner("Waiting...") as pbar:
           # The spinner starts here but no tqdm progress bar is displayed yet.
           await init_work()
           # When the user sets the 'total' attribute, the progress bar gets displayed.
           pbar.total = 10
           for i in range(10):
               await piece_of_work()
               pbar.update(1)
    """

    @staticmethod
    def alt_format_meter(
        n,
        total,
        elapsed,
        ncols=None,
        prefix="",
        ascii=False,
        unit="it",
        unit_scale=False,
        rate=None,
        bar_format=None,
        postfix=None,
        *args,
        **kwargs,
    ) -> str:
        # Return the prefix string only.
        return str(prefix) + str(postfix)

    def __init__(
        self,
        spinner_msg: str = "",
        spinner_delay: float = 0.14,
        unit: str = "it",
    ) -> None:
        self.spinner_msg = spinner_msg
        self.spinner_delay = spinner_delay
        prefix = style("", fg="bright_yellow", reset=False)
        if spinner_msg:
            initial_desc = f"{prefix}  {spinner_msg} "
        else:
            initial_desc = f"{prefix}  "
        self._orig_format_meter = self.format_meter
        super().__init__(
            total=float("inf"),
            unit=unit,
        )
        # Deactivate the progress bar display by default
        self.format_meter = self.alt_format_meter  # type: ignore
        self.set_description_str(initial_desc)
        self.set_postfix_str(style("", reset=True))

    async def spin(self) -> None:
        prefix = style("", fg="bright_yellow", reset=False)
        try:
            while True:
                for char in "|/-\\":
                    if self.spinner_msg:
                        self.set_description_str(f"{prefix}{char} {self.spinner_msg} ")
                    else:
                        self.set_description_str(f"{prefix}{char} ")
                    await asyncio.sleep(self.spinner_delay)
        except asyncio.CancelledError:
            pass

    @property
    def total(self) -> int | float | None:
        return self._total

    @total.setter
    def total(self, value: int | float) -> None:
        self._total = value
        # Reactivate the progress bar display when total is first set
        self.format_meter = self._orig_format_meter  # type: ignore

    async def __aenter__(self) -> Self:
        self.spinner_task = asyncio.create_task(self.spin())
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        if self.spinner_task is not None and not self.spinner_task.done():
            self.spinner_task.cancel()
            await self.spinner_task
        prefix = style("", fg="bright_green", reset=False)
        if self.spinner_msg:
            self.set_description_str(f"{prefix}✓ {self.spinner_msg}")
        else:
            self.set_description_str(f"{prefix}✓ ")
        self.close()
        return None



from __future__ import annotations

import enum
from typing import TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from ..config import APIConfig
    from ..output import BaseOutputHandler


class OutputMode(enum.Enum):
    CONSOLE = "console"
    JSON = "json"


@attr.define(slots=True)
class CLIContext:
    api_config: APIConfig = attr.field()
    output_mode: OutputMode = attr.field()
    output: BaseOutputHandler = attr.field(default=None)



import enum


class SubscribableEvents(enum.StrEnum):
    KERNEL_CANCELLED = "kernel_cancelled"
    KERNEL_CREATING = "kernel_creating"
    KERNEL_STARTED = "kernel_started"
    KERNEL_TERMINATED = "kernel_terminated"
    KERNEL_TERMINATING = "kernel_terminating"
    SESSION_FAILURE = "session_failure"
    SESSION_STARTED = "session_started"
    SESSION_SUCCESS = "session_success"
    SESSION_TERMINATED = "session_terminated"

    # Virtual event representing either success or failure for batch sessions
    BATCH_SESSION_RESULT = "batch_session_result"



import base64
import enum
import secrets

import attrs
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

from ai.backend.common.auth.utils import generate_signature

__all__ = (
    "AuthToken",
    "AuthTokenTypes",
    "generate_signature",
)


_iv_dict = [
    bytes([char]) for char in b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
]


class AuthTokenTypes(enum.Enum):
    KEYPAIR = "keypair"
    JWT = "jwt"


@attrs.define()
class AuthToken:
    type = attrs.field(default=AuthTokenTypes.KEYPAIR)  # type: AuthTokenTypes
    content = attrs.field(default=None)  # type: str


def encrypt_payload(endpoint: str, body: bytes) -> bytes:
    iv = b"".join(secrets.choice(_iv_dict) for _ in range(16))
    encoded_endpoint = base64.b64encode(endpoint.encode("utf-8"))
    key = (encoded_endpoint + iv + iv)[:32]
    crypt = AES.new(key, AES.MODE_CBC, iv)
    result = base64.b64encode(crypt.encrypt(pad(body, 16)))
    return iv + b":" + result


def decrypt_payload(endpoint: str, payload: bytes) -> bytes:
    iv, real_payload = payload.split(b":")
    key = (base64.b64encode(endpoint.encode("ascii")) + iv + iv)[:32]
    crypt = AES.new(key, AES.MODE_CBC, iv)
    b64p = base64.b64decode(real_payload)
    return unpad(crypt.decrypt(bytes(b64p)), 16)



import contextlib
import os
import secrets
import signal
import subprocess
import sys
from pathlib import Path
from typing import Iterator, List

from ai.backend.cli.types import ExitCode

from ..pretty import print_fail, print_info

CLI_EXECUTABLE: tuple[str, ...]
if pex_path := os.environ.get("PEX", None):
    CLI_EXECUTABLE = (sys.executable, pex_path)
else:
    CLI_EXECUTABLE = (sys.executable, "-m", "ai.backend.cli")


@contextlib.contextmanager
def container_ssh_ctx(session_ref: str, port: int) -> Iterator[Path]:
    random_id = secrets.token_hex(16)
    key_filename = "id_container"
    key_path = Path(f"~/.ssh/id_{random_id}").expanduser()
    try:
        subprocess.run(
            [*CLI_EXECUTABLE, "session", "download", session_ref, key_filename],
            shell=False,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError as e:
        print_fail(f"Failed to download the SSH key from the session (exit: {e.returncode}):")
        print(e.stdout.decode())
        sys.exit(ExitCode.FAILURE)
    os.rename(key_filename, key_path)
    print_info(f"running a temporary sshd proxy at localhost:{port} ...", file=sys.stderr)
    # proxy_proc is a background process
    proxy_proc = subprocess.Popen(
        [*CLI_EXECUTABLE, "app", session_ref, "sshd", "-b", f"127.0.0.1:{port}"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    assert proxy_proc.stdout is not None
    try:
        lines: List[bytes] = []
        while True:
            line = proxy_proc.stdout.readline(1024)
            if not line:
                proxy_proc.wait()
                print_fail(
                    "Unexpected early termination of the sshd app command "
                    f"(exit: {proxy_proc.returncode}):"
                )
                print((b"\n".join(lines)).decode())
                sys.exit(ExitCode.FAILURE)
            if f"127.0.0.1:{port}".encode() in line:
                break
            lines.append(line)
        lines.clear()
        yield key_path
    finally:
        proxy_proc.send_signal(signal.SIGINT)
        proxy_proc.wait()
        os.unlink(key_path)



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import Any, Mapping, Optional, Sequence, TypeAlias, TypeVar, Union
from uuid import UUID

import aiohttp
import janus
from aiohttp import hdrs
from aiotusclient import client
from tenacity import (
    AsyncRetrying,
    RetryError,
    TryAgain,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from tqdm import tqdm
from yarl import URL

from ai.backend.client.output.fields import vfolder_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult
from ai.backend.common.types import ResultSet

from ..compat import current_loop
from ..config import DEFAULT_CHUNK_SIZE, MAX_INFLIGHT_CHUNKS
from ..exceptions import BackendClientError
from ..pagination import fetch_paginated_result
from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("VFolderByName",)

_default_list_fields = (
    vfolder_fields["host"],
    vfolder_fields["name"],
    vfolder_fields["status"],
    vfolder_fields["created_at"],
    vfolder_fields["creator"],
    vfolder_fields["group_id"],
    vfolder_fields["permission"],
    vfolder_fields["ownership_type"],
)

T = TypeVar("T")
list_: TypeAlias = list[T]


class ResponseFailed(Exception):
    pass


class VFolderByName(BaseFunction):
    name: str
    id: Optional[UUID] = None

    def __init__(self, name: str, id: Optional[UUID] = None) -> None:
        self.name = name
        self.id = id

    @api_function
    @classmethod
    async def create(
        cls,
        name: str,
        host: Optional[str] = None,
        unmanaged_path: Optional[str] = None,
        group: Optional[str] = None,
        usage_mode: str = "general",
        permission: str = "rw",
        cloneable: bool = False,
    ):
        rqst = Request("POST", "/folders")
        rqst.set_json({
            "name": name,
            "host": host,
            "unmanaged_path": unmanaged_path,
            "group": group,
            "usage_mode": usage_mode,
            "permission": permission,
            "cloneable": cloneable,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def delete_by_id(cls, oid):
        rqst = Request("DELETE", "/folders")
        rqst.set_json({"id": oid})
        async with rqst.fetch():
            return {}

    @api_function
    @classmethod
    async def list(cls, list_all=False):
        rqst = Request("GET", "/folders")
        rqst.set_json({"all": list_all})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        group: Optional[str] = None,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> PaginatedResult[dict]:
        """
        Fetches the list of vfolders. Domain admins can only get domain vfolders.

        :param group: Fetch vfolders in a specific group.
        :param fields: Additional per-vfolder query fields to fetch.
        """
        return await fetch_paginated_result(
            "vfolder_list",
            {
                "group_id": (group, "UUID"),
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def paginated_own_list(
        cls,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> PaginatedResult[dict]:
        """
        Fetches the list of own vfolders.

        :param fields: Additional per-vfolder query fields to fetch.
        """
        return await fetch_paginated_result(
            "vfolder_own_list",
            {
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def paginated_invited_list(
        cls,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> PaginatedResult[dict]:
        """
        Fetches the list of invited vfolders.

        :param fields: Additional per-vfolder query fields to fetch.
        """
        return await fetch_paginated_result(
            "vfolder_invited_list",
            {
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def paginated_project_list(
        cls,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> PaginatedResult[dict]:
        """
        Fetches the list of invited vfolders.

        :param fields: Additional per-vfolder query fields to fetch.
        """
        return await fetch_paginated_result(
            "vfolder_project_list",
            {
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    async def _get_id_by_name(self) -> UUID:
        rqst = Request("GET", "/folders/_/id")
        rqst.set_json({
            "name": self.name,
        })
        async with rqst.fetch() as resp:
            data = await resp.json()
            return UUID(data["id"])

    @api_function
    async def get_id(self) -> UUID:
        return await self._get_id_by_name()

    @api_function
    @classmethod
    async def list_hosts(cls):
        rqst = Request("GET", "/folders/_/hosts")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_all_hosts(cls):
        rqst = Request("GET", "/folders/_/all_hosts")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_allowed_types(cls):
        rqst = Request("GET", "/folders/_/allowed_types")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def info(self):
        await self.update_id_by_name()
        rqst = Request("GET", "/folders/{0}".format(self.request_key))
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete(self):
        await self.update_id_by_name()
        rqst = Request("DELETE", "/folders")
        rqst.set_json({
            "id": self.request_key,
        })
        async with rqst.fetch():
            return {}

    @api_function
    async def purge(self) -> Mapping[str, Any]:
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/purge")
        rqst.set_json({
            "id": self.request_key,
        })
        async with rqst.fetch():
            return {}

    async def _restore(self) -> Mapping[str, Any]:
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/restore-from-trash-bin")
        rqst.set_json({
            "id": self.request_key,
        })
        async with rqst.fetch():
            return {}

    @api_function
    async def recover(self):
        return await self._restore()

    @api_function
    async def restore(self):
        return await self._restore()

    @api_function
    async def delete_trash(self) -> Mapping[str, Any]:
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/delete-from-trash-bin")
        rqst.set_json({
            "id": self.request_key,
        })
        async with rqst.fetch():
            return {}

    @api_function
    async def force_delete(self) -> dict[str, Any]:
        await self.update_id_by_name()
        rqst = Request("DELETE", "/folders/{0}/force".format(self.request_key))
        async with rqst.fetch():
            return {}

    @api_function
    async def rename(self, new_name):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{0}/rename".format(self.request_key))
        rqst.set_json({
            "new_name": new_name,
        })
        async with rqst.fetch() as resp:
            self.name = new_name
            return await resp.text()

    def _write_file(self, file_path: Path, mode: str, q: janus._SyncQueueProxy[bytes]):
        with open(file_path, mode) as f:
            while True:
                chunk = q.get()
                if not chunk:
                    return
                f.write(chunk)
                q.task_done()

    async def _download_file(
        self,
        file_path: Path,
        download_url: URL,
        chunk_size: int,
        max_retries: int,
        show_progress: bool,
    ) -> None:
        if show_progress:
            print(f"Downloading to {file_path} ...")

        range_start = 0
        if_range: str | None = None
        file_unit = "bytes"
        file_mode = "wb"
        file_req_hdrs: dict[str, str] = {}
        try:
            async for session_attempt in AsyncRetrying(
                wait=wait_exponential(multiplier=0.02, min=0.02, max=5.0),
                stop=stop_after_attempt(max_retries),
                retry=retry_if_exception_type(TryAgain),
            ):
                with session_attempt:
                    try:
                        if if_range is not None:
                            file_req_hdrs[hdrs.IF_RANGE] = if_range
                            file_req_hdrs[hdrs.RANGE] = f"{file_unit}={range_start}-"
                        async with aiohttp.ClientSession(headers=file_req_hdrs) as client:
                            async with client.get(download_url, ssl=False) as raw_resp:
                                match raw_resp.status:
                                    case 200:
                                        # First attempt to download file or file has changed.
                                        file_mode = "wb"
                                        range_start = 0
                                    case 206:
                                        # File has not changed. Continue downloading from range_start.
                                        file_mode = "ab"
                                    case _:
                                        # Retry.
                                        raise ResponseFailed
                                size = int(raw_resp.headers["Content-Length"])
                                if_range = raw_resp.headers.get("Last-Modified")
                                q: janus.Queue[bytes] = janus.Queue(MAX_INFLIGHT_CHUNKS)
                                try:
                                    with tqdm(
                                        total=(size - range_start),
                                        unit=file_unit,
                                        unit_scale=True,
                                        unit_divisor=1024,
                                        disable=not show_progress,
                                    ) as pbar:
                                        loop = current_loop()
                                        writer_fut = loop.run_in_executor(
                                            None,
                                            self._write_file,
                                            file_path,
                                            file_mode,
                                            q.sync_q,
                                        )
                                        await asyncio.sleep(0)
                                        max_attempts = 10
                                        while True:
                                            try:
                                                async for attempt in AsyncRetrying(
                                                    wait=wait_exponential(
                                                        multiplier=0.02,
                                                        min=0.02,
                                                        max=5.0,
                                                    ),
                                                    stop=stop_after_attempt(max_attempts),
                                                    retry=retry_if_exception_type(TryAgain),
                                                ):
                                                    with attempt:
                                                        try:
                                                            chunk = await raw_resp.content.read(
                                                                chunk_size
                                                            )
                                                        except asyncio.TimeoutError:
                                                            raise TryAgain
                                            except RetryError:
                                                raise ResponseFailed
                                            range_start += len(chunk)
                                            pbar.update(len(chunk))
                                            if not chunk:
                                                break
                                            await q.async_q.put(chunk)
                                finally:
                                    await q.async_q.put(b"")
                                    await writer_fut
                                    q.close()
                                    await q.wait_closed()
                    except (
                        ResponseFailed,
                        aiohttp.ClientPayloadError,
                        aiohttp.ClientConnectorError,
                    ):
                        raise TryAgain
        except RetryError:
            raise RuntimeError(f"Downloading {file_path.name} failed after {max_retries} retries")

    @api_function
    async def download(
        self,
        relative_paths: Sequence[Union[str, Path]],
        *,
        basedir: Optional[Union[str, Path]] = None,
        dst_dir: Optional[Union[str, Path]] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        show_progress: bool = False,
        address_map: Optional[Mapping[str, str]] = None,
        max_retries: int = 20,
    ) -> None:
        base_path = Path.cwd() if basedir is None else Path(basedir).resolve()
        await self.update_id_by_name()
        for relpath in relative_paths:
            file_path = base_path / relpath
            if file_path.exists():
                raise RuntimeError("The target file already exists", file_path.name)

            rqst = Request("POST", "/folders/{}/request-download".format(self.request_key))
            rqst.set_json({
                "path": str(relpath),
            })
            async with rqst.fetch() as resp:
                download_info = await resp.json()
                overriden_url = download_info["url"]
                if address_map:
                    if download_info["url"] in address_map:
                        overriden_url = address_map[download_info["url"]]
                    else:
                        raise BackendClientError(
                            "Overriding storage proxy addresses are given, "
                            "but no url matches with any of them.\n",
                        )

                params = {"token": download_info["token"]}
                if dst_dir is not None:
                    params["dst_dir"] = dst_dir
                download_url = URL(overriden_url).with_query(params)
            await self._download_file(
                file_path, download_url, chunk_size, max_retries, show_progress
            )

    async def _upload_files(
        self,
        file_paths: Sequence[Path],
        basedir: Optional[Union[str, Path]] = None,
        dst_dir: Optional[Union[str, Path]] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        address_map: Optional[Mapping[str, str]] = None,
    ) -> None:
        base_path = Path.cwd() if basedir is None else Path(basedir).resolve()
        await self.update_id_by_name()
        for file_path in file_paths:
            if file_path.is_dir():
                raise BackendClientError(
                    f"Failed to upload {file_path}. Use recursive option to upload directories."
                )
            file_size = Path(file_path).stat().st_size
            rqst = Request("POST", "/folders/{}/request-upload".format(self.request_key))
            rqst.set_json({
                "path": "{}".format(str(Path(file_path).relative_to(base_path))),
                "size": int(file_size),
            })
            async with rqst.fetch() as resp:
                upload_info = await resp.json()
                overriden_url = upload_info["url"]
                if address_map:
                    if upload_info["url"] in address_map:
                        overriden_url = address_map[upload_info["url"]]
                    else:
                        raise BackendClientError(
                            "Overriding storage proxy addresses are given, "
                            "but no url matches with any of them.\n",
                        )
                params = {"token": upload_info["token"]}
                if dst_dir is not None:
                    params["dst_dir"] = dst_dir
                upload_url = URL(overriden_url).with_query(params)
            tus_client = client.TusClient()
            if basedir:
                input_file = open(base_path / file_path, "rb")
            else:
                input_file = open(str(Path(file_path).relative_to(base_path)), "rb")
            print(f"Uploading {base_path / file_path} via {upload_info['url']} ...")
            # TODO: refactor out the progress bar
            uploader = tus_client.async_uploader(
                file_stream=input_file,
                url=upload_url,
                upload_checksum=False,
                chunk_size=chunk_size,
            )
            await uploader.upload()
            input_file.close()

    async def _upload_recursively(
        self,
        source: Sequence[Path],
        basedir: Optional[Union[str, Path]] = None,
        dst_dir: Optional[Union[str, Path]] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        address_map: Optional[Mapping[str, str]] = None,
    ) -> None:
        dir_list: list[Path] = []
        file_list: list[Path] = []
        base_path = Path.cwd() if basedir is None else Path(basedir).resolve()
        for path in source:
            if path.is_file():
                file_list.append(path)
            else:
                await self._mkdir([path.relative_to(base_path)])
                dir_list.append(path)
        await self._upload_files(file_list, basedir, dst_dir, chunk_size, address_map)
        for dir in dir_list:
            await self._upload_recursively(
                list(dir.glob("*")), basedir, dst_dir, chunk_size, address_map
            )

    @api_function
    async def upload(
        self,
        sources: Sequence[Union[str, Path]],
        *,
        basedir: Optional[Union[str, Path]] = None,
        recursive: bool = False,
        dst_dir: Optional[Union[str, Path]] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        address_map: Optional[Mapping[str, str]] = None,
        show_progress: bool = False,
    ) -> None:
        if basedir:
            src_paths = [basedir / Path(src) for src in sources]
        else:
            src_paths = [Path(src).resolve() for src in sources]
        if recursive:
            await self._upload_recursively(src_paths, basedir, dst_dir, chunk_size, address_map)
        else:
            await self._upload_files(src_paths, basedir, dst_dir, chunk_size, address_map)

    async def _mkdir(
        self,
        path: str | Path | list_[str | Path],
        parents: Optional[bool] = False,
        exist_ok: Optional[bool] = False,
    ) -> ResultSet:
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/mkdir".format(self.request_key))
        rqst.set_json({
            "path": path,
            "parents": parents,
            "exist_ok": exist_ok,
        })
        async with rqst.fetch() as resp:
            reply = await resp.json()
            return reply["results"]

    @api_function
    async def mkdir(
        self,
        path: str | Path | list_[str | Path],
        parents: Optional[bool] = False,
        exist_ok: Optional[bool] = False,
    ) -> ResultSet:
        return await self._mkdir(path, parents, exist_ok)

    @api_function
    async def rename_file(self, target_path: str, new_name: str):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/rename-file".format(self.request_key))
        rqst.set_json({
            "target_path": target_path,
            "new_name": new_name,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def move_file(self, src_path: str, dst_path: str):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/move-file".format(self.request_key))
        rqst.set_json({
            "src": src_path,
            "dst": dst_path,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete_files(self, files: Sequence[Union[str, Path]], recursive: bool = False):
        await self.update_id_by_name()
        rqst = Request("DELETE", "/folders/{}/delete-files".format(self.request_key))
        rqst.set_json({
            "files": files,
            "recursive": recursive,
        })
        async with rqst.fetch() as resp:
            return await resp.text()

    @api_function
    async def list_files(self, path: Union[str, Path] = "."):
        await self.update_id_by_name()
        rqst = Request(
            "GET", "/folders/{}/files".format(self.request_key), params={"path": str(path)}
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def invite(self, perm: str, emails: Sequence[str]):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/invite".format(self.request_key))
        rqst.set_json({
            "perm": perm,
            "user_ids": emails,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def invitations(cls):
        rqst = Request("GET", "/folders/invitations/list")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def accept_invitation(cls, inv_id: str):
        rqst = Request("POST", "/folders/invitations/accept")
        rqst.set_json({"inv_id": inv_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def delete_invitation(cls, inv_id: str):
        rqst = Request("DELETE", "/folders/invitations/delete")
        rqst.set_json({"inv_id": inv_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_fstab_contents(cls, agent_id=None):
        rqst = Request(
            "GET",
            "/folders/_/fstab",
            params={
                "agent_id": agent_id,
            },
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_performance_metric(cls, folder_host: str):
        rqst = Request("GET", "/folders/_/perf-metric", params={"folder_host": folder_host})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_mounts(cls):
        rqst = Request("GET", "/folders/_/mounts")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def mount_host(cls, name: str, fs_location: str, options=None, edit_fstab: bool = False):
        rqst = Request("POST", "/folders/_/mounts")
        rqst.set_json({
            "name": name,
            "fs_location": fs_location,
            "options": options,
            "edit_fstab": edit_fstab,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def umount_host(cls, name: str, edit_fstab: bool = False):
        rqst = Request("DELETE", "/folders/_/mounts")
        rqst.set_json({
            "name": name,
            "edit_fstab": edit_fstab,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def share(self, perm: str, emails: Sequence[str]):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/share".format(self.request_key))
        rqst.set_json({
            "permission": perm,
            "emails": emails,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def unshare(self, emails: Sequence[str]):
        await self.update_id_by_name()
        rqst = Request("DELETE", "/folders/{}/unshare".format(self.request_key))
        rqst.set_json({
            "emails": emails,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def leave(self, shared_user_uuid=None):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/leave".format(self.request_key))
        rqst.set_json({"shared_user_uuid": shared_user_uuid})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def clone(
        self,
        target_name: str,
        target_host: Optional[str] = None,
        usage_mode: str = "general",
        permission: str = "rw",
    ):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/clone".format(self.request_key))
        rqst.set_json({
            "target_name": target_name,
            "target_host": target_host,
            "usage_mode": usage_mode,
            "permission": permission,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def update_options(
        self, name: str, permission: Optional[str] = None, cloneable: Optional[bool] = None
    ):
        await self.update_id_by_name()
        rqst = Request("POST", "/folders/{}/update-options".format(self.request_key))
        rqst.set_json({
            "cloneable": cloneable,
            "permission": permission,
        })
        async with rqst.fetch() as resp:
            return await resp.text()

    @api_function
    @classmethod
    async def list_shared_vfolders(cls):
        rqst = Request("GET", "folders/_/shared")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def shared_vfolder_info(cls, vfolder_id: str):
        rqst = Request("GET", "folders/_/shared", params={"vfolder_id": vfolder_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def update_shared_vfolder(cls, vfolder: str, user: str, perm: Optional[str] = None):
        rqst = Request("POST", "/folders/_/shared")
        rqst.set_json({
            "vfolder": vfolder,
            "user": user,
            "perm": perm,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def change_vfolder_ownership(cls, vfolder: str, user_email: str):
        rqst = Request("POST", "/folders/_/change-ownership")
        rqst.set_json({
            "vfolder": vfolder,
            "user_email": user_email,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    async def update_id_by_name(self) -> None:
        if self.id is None:
            self.id = await self._get_id_by_name()

    @property
    def request_key(self) -> str:
        if self.id is not None:
            return self.id.hex
        return self.name



import json
from typing import Any, Optional

from ..auth import encrypt_payload
from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("Auth",)


def _put_secure_body(rqst: Request, data: Any) -> None:
    if rqst.config.endpoint.scheme == "https":
        rqst.set_json(data)
    else:
        rqst.headers["X-BackendAI-Encoded"] = "true"
        raw_body = json.dumps(data).encode()
        encoded_body = encrypt_payload(str(rqst.config.endpoint), raw_body)
        rqst.set_content(encoded_body)


class Auth(BaseFunction):
    """
    Provides the function interface for login session management and authorization.
    """

    @api_function
    @classmethod
    async def login(cls, user_id: str, password: str, otp: Optional[str] = None) -> dict:
        """
        Log-in into the endpoint with the given user ID and password.
        It creates a server-side web session and return
        a dictionary with ``"authenticated"`` boolean field and
        JSON-encoded raw cookie data.

        This SDK function works when env var BACKEND_ENDPOINT is endpoint of webserver, not manager.
        If you want to use SDK funcs that send requests to the Manager with login,
        update the cookie in the session context manager object and save the state with a file.
        See the example in `src/ai/backend/client/cli/config.py` login function.
        """
        rqst = Request("POST", "/server/login")
        body = {
            "username": user_id,
            "password": password,
        }
        if otp:
            body["otp"] = otp
        _put_secure_body(rqst, body)
        async with rqst.fetch(anonymous=True) as resp:
            data = await resp.json()
            data["cookies"] = resp.raw_response.cookies
            data["config"] = {
                "username": user_id,
            }
            return data

    @api_function
    @classmethod
    async def logout(cls) -> None:
        """
        Log-out from the endpoint.
        It clears the server-side web session.

        This SDK function works when env var BACKEND_ENDPOINT is endpoint of webserver, not manager.
        """
        rqst = Request("POST", "/server/logout")
        async with rqst.fetch() as resp:
            resp.raw_response.raise_for_status()

    @api_function
    @classmethod
    async def update_password(
        cls, old_password: str, new_password: str, new_password2: str
    ) -> dict:
        """
        Update user's password. This API works only for account owner.
        """
        rqst = Request("POST", "/auth/update-password")
        body = {
            "old_password": old_password,
            "new_password": new_password,
            "new_password2": new_password2,
        }
        _put_secure_body(rqst, body)
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def update_password_no_auth(
        cls, domain: str, user_id: str, current_password: str, new_password: str
    ) -> dict:
        """
        Update user's password. This is used to update `EXPIRED` password only.
        This function fetch a request to manager.
        """

        rqst = Request("POST", "/auth/update-password-no-auth")
        body = {
            "domain": domain,
            "username": user_id,
            "current_password": current_password,
            "new_password": new_password,
        }
        _put_secure_body(rqst, body)
        async with rqst.fetch(anonymous=True) as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def update_password_no_auth_in_session(
        cls, user_id: str, current_password: str, new_password: str
    ) -> dict:
        """
        Update user's password. This is used to update `EXPIRED` password only.
        This function fetch a request to webserver.
        """

        rqst = Request("POST", "/server/update-password-no-auth")
        body = {
            "username": user_id,
            "current_password": current_password,
            "new_password": new_password,
        }
        _put_secure_body(rqst, body)
        async with rqst.fetch(anonymous=True) as resp:
            return await resp.json()
